#include <fcntl.h>  // O_RDONLY, etc..
#include <unistd.h>
#include <stdlib.h> // exit(), etc..
#include <string.h> // strncmp, etc.
#include <stdio.h>  // printf(), etc..
#include <sys/stat.h> // mkdir(), stat()...
#include <sys/mman.h> // mmap(), etc
#include <zlib.h>
#ifdef HAVE_LIBSPARSE
#include "sparse_file.h"
#endif
#ifdef HAVE_SHA
#include "sha.h"
#endif



#define VERSION	"0.4"
/**
  *
  * Rudimentary Android image and partition unpacking tool
  *
  * v0.1 - Support unpacking of BootLoader and Boot images, as well as imgdata
  * v0.2 - Supports offset= (for HTC and other boot.imgs where ANDROID! is wrapped)
  *        Supports cmdline= (to override kernel command line)
  *
  * v0.3 - Integrated mkbootimg functionality, added dt_size and id
  *
  * v0.4 - cmdline, addrs..
  * 
  * @TODO: IMGDATA
  * @TODO: cache/userdata/system.img support - via simg
  *
  *
  * This tool is part of the free downloads for the book "Android Internals: A Confectioner's Cookbook"
  *
  * Author: Jonathan Levin, http://NewAndroidBook.com
  *
  **/


#define GZ_MAGIC "\x1f\x8b\x08\x00\x00\x00\x00\x00"
#define LZO_MAGIC "\x89\x4c\x5a\x4f\x00\x0d\x0a\x1a\x0a"
//#define LZO_MAGIC "\x4c\x5a\x4f\x00\x0d\x0a\x1a\x0a"


#define SIMG_MAGIC	0xed26ff3a

struct simg_header
{

	uint32_t	magic;
	uint16_t	version_major;
	uint16_t	version_minor;
	uint16_t	header_size;
	uint16_t	chunk_size;
	uint32_t	block_size;
	uint32_t	num_blocks;
	uint32_t	num_chunks;
	uint32_t	checksum;


};

// From releasetools.pl:

#define BOOTLDR_MAGIC "BOOTLDR!"
#define BOOTLDR_MAGIC_SIZE 8

struct bootloader_images_header {
         char magic[BOOTLDR_MAGIC_SIZE];
         unsigned int num_images;
         unsigned int start_offset;
         unsigned int bootldr_size;
         struct {
                 char name[64];
                 unsigned int size;
         } img_info[];
};



// Verbatim copy from bootimg.h

typedef struct boot_img_hdr boot_img_hdr;

#define BOOT_MAGIC "ANDROID!"
#define BOOT_MAGIC_SIZE 8
#define BOOT_NAME_SIZE 16
#define BOOT_ARGS_SIZE 512


// q.v. https://www.codeaurora.org/cgit/quic/la/kernel/lk/tree/app/aboot/bootimg.h
struct boot_img_hdr
{
    unsigned char magic[BOOT_MAGIC_SIZE];

    unsigned kernel_size;  /* size in bytes */
    unsigned kernel_addr;  /* physical load addr */

    unsigned ramdisk_size; /* size in bytes */
    unsigned ramdisk_addr; /* physical load addr */

    unsigned second_size;  /* size in bytes */
    unsigned second_addr;  /* physical load addr */

    unsigned tags_addr;    /* physical addr for kernel tags */
    unsigned page_size;    /* flash mmapped size we assume */
    unsigned dt_size;	   /* In newer versions of bootimg.h */
		// q.v bootimg.h?h=master&id=5e68eb8a09820d8447cb209558fbc237fcbfc41f

    unsigned unused[1];    /* future expansion: should be 0 */

    unsigned char name[BOOT_NAME_SIZE]; /* asciiz product name */
    
    unsigned char cmdline[BOOT_ARGS_SIZE];

    unsigned id[8]; /* timestamp / checksum / sha1 / etc */
};

/*
** +-----------------+ 
** | boot header     | 1 page 
** +-----------------+
** | kernel          | n pages  
** +-----------------+
** | ramdisk         | m pages  
** +-----------------+
** | second stage    | o pages
** +-----------------+
**
** n = (kernel_size + page_size - 1) / page_size
** m = (ramdisk_size + page_size - 1) / page_size
** o = (second_size + page_size - 1) / page_size
**
** 0. all entities are page_size aligned in flash
** 1. kernel and ramdisk are required (size != 0)
** 2. second is optional (second_size == 0 -> no second)
** 3. load each element (kernel, ramdisk, second) at
**    the specified physical address (kernel_addr, etc)
** 4. prepare tags at tag_addr.  kernel_args[] is
**    appended to the kernel commandline in the tags.
** 5. r0 = 0, r1 = MACHINE_TYPE, r2 = tags_addr
** 6. if second_size != 0: jump to second_addr
**    else: jump to kernel_addr
*/

#if 0
typedef struct ptentry ptentry;

struct ptentry {
    char name[16];      /* asciiz partition name    */
    unsigned start;     /* starting block number    */
    unsigned length;    /* length in blocks         */
    unsigned flags;     /* set to zero              */
};

/* MSM Partition Table ATAG
**
** length: 2 + 7 * n
** atag:   0x4d534d70
**         <ptentry> x n
*/
#endif


// From reverse engineering

#define IMGDATA_MAGIC "IMGDATA!"
struct imgdata_header {
         char magic[BOOTLDR_MAGIC_SIZE];
	 unsigned int unknown;
         unsigned int num_images;
         unsigned int unknown1;
         unsigned int unknown2;
         struct {
                 char name[16];
                 unsigned int width;
		 unsigned int height;
		 unsigned int unknown3;
		 unsigned int unknown4;
		 unsigned int flash_offset;
		 unsigned int flash_size;
         } img_info[];
};


// Let's get to work

#define EXTRACTED_DIR	"extracted"

#define true	1
#define	false	0

#ifdef HAVE_LIBSPARSE
int extractSimg (int in)
{
   mkdir("extracted", 0777);
    int out = open ("extracted/image.img", O_WRONLY | O_CREAT);

  if (out < 0 ) {perror ("extracted/image.img"); exit(1);}
  fchmod(out, 0644);

struct sparse_file *s = sparse_file_import(in, true, false);
                if (!s) {
                        fprintf(stderr, "Failed to read sparse file\n");
                        exit(-1);
                }

                lseek(out, SEEK_SET, 0);

                int ret = sparse_file_write(s, out, false, false, false);
                if (ret < 0) {
                        fprintf(stderr, "Cannot write output file\n");
                        exit(-1);
                }
                sparse_file_destroy(s);
                close(in);

  printf("Extracted image is in extracted/image.img\n");
  return (0);

}
#else
int extractSimg(int in)
{
	printf("This version of imgtool was compiled without libsparse - can't extract simg files\n");
}
#endif

int decompressKernel (unsigned char *kernel, int size)
{

    int i = 0;

    int gz =0, lzo = 0;

    for (i = 0; i < size; i++)
        {

                if (memcmp(kernel +i, GZ_MAGIC, 8) == 0)
                {
                        fprintf(stderr, "Found GZ Magic at offset %d\n",i);
			gz++;
                        break;
                }

		if (memcmp(kernel +i, LZO_MAGIC, 4) == 0)
		{

                        fprintf(stderr, "Found LZO Magic at offset %d\n",i);
			lzo++;
                        break;
		
		}
        } // end for


   // Quick & Dirty -
   // This version doesn't have GZip support, so instead we write out the compressed portion of the file, 
   // and then invoke gunzip - because we have the magic, it will take care of that.

    int outputFd = -1;
    int dtFd = -1;
    system ("mkdir ./extracted 2> /dev/null");
    if (gz)  outputFd  = open ( "extracted/kernelimage.gz", O_WRONLY | O_CREAT | O_TRUNC);
    if (lzo)  outputFd = open ( "extracted/kernelimage.lzo", O_WRONLY | O_CREAT | O_TRUNC);
    if (! gz && !lzo) { fprintf(stderr,"Can't find either GZ or LZO signature here... won't decompress\n"); return 1;

	}
    if (outputFd == -1) { perror ("open"); exit(9);}
    fchmod(outputFd, 0600);
    write (outputFd, kernel+i, size -i);


    fprintf(stderr, "Looking for device tree...\n");

    // Known Bug: There may be more than one device tree. This only finds the first one.
    // Yeah. I know.

#define DT_MAGIC 0xd00dfeed

    int j = 0;
    int dt_magic = DT_MAGIC;
    for (j =  i + 7000000;  j < i+ size; j++)
	{
	 
		if ( (kernel[j] == 0xd0) && (kernel[j+1] == 0x0d) &&
		    (kernel[j+2] == 0xfe) && (kernel[j+3] == 0xed))
		{
			fprintf(stderr, "Found DT Magic @%x\n", j);
			dtFd = open ("extracted/devicetree.dtb", O_WRONLY | O_CREAT | O_TRUNC);
			write (dtFd, kernel + j , size - j);
			fchmod (dtFd, 0600);
			close (dtFd);
			break;
		}
	}
    close (outputFd);

 
   // The *easiest* way to uncompress a kernel - no sense bother with zlib when gunzip is a given.
   if (gz) {
   if (access("/usr/bin/gunzip", X_OK) == 0) { system("/usr/bin/gunzip -v extracted/kernelimage.gz"); return 0;}
   if (access("/bin/gunzip", X_OK) == 0) { system("/bin/gunzip -v extracted/kernelimage.gz"); return 0;} 
	fprintf(stderr, "This version of imgtool relies on gunzip to do extraction, but gunzip wasn't found\n");	
	return 1;
        
	}



  return 0;
}



void extractPortion (char *mmapped, 
		     int   offset,
		     int   size,
		     char *name)
{
	
	// extract a portion of a memory mapped file, mmapped, size bytes from offset. No bounds checking,
	// No nothin'.

	fprintf(stderr,"extracting %s\n", name);
	mkdir (EXTRACTED_DIR,0777);
	chdir (EXTRACTED_DIR);
	int fd = open (name, O_WRONLY | O_TRUNC | O_CREAT);
	fchmod (fd, 0666);  // yeah, permissions...
	if (fd < 0) {
		perror ("name");
		return;

	}
	write (fd, mmapped + offset, size);
	close (fd);

	chdir ("..");


}


char *makeImg (char *imgName, char *kernelName, char *ramdiskName, uint32_t Addr, char *CmdLine)
{

	int k = open (kernelName, O_RDONLY);
	if (k < 0) { perror (kernelName); return 1001; }

	struct stat stbuf;
	int rc = fstat(k, &stbuf);
	int kernelSize = stbuf.st_size;


	int rd = open (ramdiskName, O_RDONLY);
	if (rd < 0) { perror (ramdiskName); return 1002; }
	rc = fstat(rd, &stbuf);
	int ramdiskSize = stbuf.st_size;


	char *mmappedKernel = mmap(NULL,
                        kernelSize,
                        PROT_READ,
                        MAP_PRIVATE,
                        k,
                        0);

	char *mmappedRamdisk = mmap(NULL,
                        ramdiskSize,
                        PROT_READ,
                        MAP_PRIVATE,
                        rd,
                        0);


	int out = open (imgName, O_CREAT | O_TRUNC | O_WRONLY);


	int doHeader = 0;

	if (doHeader)
	{

	 	struct extra_header {
			uint32_t	magic;
			uint32_t	zeros;
			uint32_t	total_size;
			uint32_t	_0x100;
			uint32_t	total_size_minus_0x200;
			uint32_t	_0x100_here;
			uint32_t	total_size_minus_0x100;
			uint32_t	_0x100_here_too;
			uint32_t	padding[256];


		} extra ;

		memset (&extra, '\0', sizeof(extra));
		extra.magic = 0x12345678;
		extra.total_size = 8543800;
		extra.total_size_minus_0x200 = extra.total_size - 0x200;
		extra.total_size_minus_0x100 = extra.total_size - 0x100;

	write (out, &extra, 256);
		
	}
	struct boot_img_hdr bih;
	memset (&bih, '\0', sizeof(bih));

	strncpy (bih.magic, BOOT_MAGIC, BOOT_MAGIC_SIZE);
	bih.kernel_size = kernelSize;
	bih.ramdisk_size = ramdiskSize;
	bih.second_size = bih.second_addr = 0;
	bih.page_size	= 2048;

	if (!Addr) Addr = 0x10008000;
	bih.kernel_addr  = Addr;

	bih.ramdisk_addr = Addr + 0x04000000; // (ramdiskSize/ bih.page_size);
	bih.second_addr  = 0x80f00000;
	bih.tags_addr = 0x8e000000;




	// Get Sha-1
#ifdef HAVE_SHA
	SHA_CTX ctx;
	const uint8_t* sha;


        SHA_init(&ctx);
    	SHA_update(&ctx, mmappedKernel, bih.kernel_size);
        SHA_update(&ctx, &bih.kernel_size, sizeof(bih.kernel_size));
        SHA_update(&ctx, mmappedRamdisk, bih.ramdisk_size);
        SHA_update(&ctx, &bih.ramdisk_size, sizeof(bih.ramdisk_size));
        SHA_update(&ctx, "", bih.second_size);
        SHA_update(&ctx, &bih.second_size, sizeof(bih.second_size));
        sha = SHA_final(&ctx);
        memcpy(bih.id, sha,
           SHA_DIGEST_SIZE > sizeof(bih.id) ? sizeof(bih.id) : SHA_DIGEST_SIZE);

#endif


	strcpy (bih.name, "1490613711");
	memset (bih.cmdline, '\0', sizeof(bih.cmdline));
	if (CmdLine)
	{
		strncpy(bih.cmdline, (unsigned char *) CmdLine,BOOT_ARGS_SIZE-1);
	}
	else
	strcpy (bih.cmdline, (unsigned char *) "androidboot.hardware=flounder androidboot.selinux=disabled");

	write (out, &bih, sizeof(bih));

	int padding =  bih.page_size - sizeof(bih);

	char *pad = calloc(padding,1);
	write (out, pad, padding);
	free(pad);

	write (out, mmappedKernel, kernelSize);

	// Write the Ramdisk - but not so fast - must calculate this on a flash page size
	// integer boundary
	
	int whereWeAre = (sizeof(bih) + padding +  kernelSize);

	padding =  bih.page_size - ( whereWeAre % bih.page_size);

	pad = calloc(padding,1);
	write (out, pad, padding);
	free(pad);

	write (out, mmappedRamdisk, ramdiskSize);

	// And let's not forget the end padding, either..
	whereWeAre += padding + ramdiskSize;

	padding = bih.page_size - (whereWeAre % bih.page_size);
	pad = calloc(padding,1);
	write (out, pad, padding);
	free(pad);

	fchmod (out, 0600);


	close(out);

	return 0;

}

int main (int argc, char **argv)
{

	
	if (argc < 2)
	{
		fprintf (stderr, "Usage: %s _img_name_  [extract]\n",argv[0]);
		fprintf (stderr, "Where: _img_name_ is the name of an Android boot or bootloader image (or boot partition)\n");
		fprintf (stderr,"       [extract]  is an optional parameter to extract the image components\n");
		fprintf (stderr,"       [offset=...]  offset to find ANDROID! magic (e.g. 256 in HTC boot)\n");
		fprintf (stderr,"\nor:     %s make _img_name_ _kernel_ _ramdisk_ [....]\n", argv[0]);
		fprintf (stderr,"        Make _img_name by combining kernel and ramdisk and creating header\n");
		fprintf (stderr,"       [cmdline='args to kernel'] is an optional parameter specifying the kernel command line\n");
		fprintf(stderr, "       [addr=0x.....] is an optional base address to load the kernel into\n");
    
	
		fprintf(stderr,"\n\nThis is ImgTool " VERSION " compiled on " __DATE__ "\n");
		exit(0);
	}

	if (strcmp(argv[1],"make") == 0)
	{
		char *addr =NULL;
		char *cmdline = NULL;
		if (argc <5)
		{
			fprintf (stderr,"Make: Requires at least three arguments - image name, kernel and ramdisk\n");
			exit(1);
		}
		else
		{
			int i = 0;
			for (i = 4; i < argc; i++)
			{
	  		if (strstr(argv[i], "cmdline=")) cmdline = (strstr(argv[i],"=") + 1);
	  if (strstr(argv[i], "addr=")) addr = (strstr(argv[i],"=") + 1);
			}
			// Assume argv[2-4] are image Name, kernel, ramdisk
			
		}
		uint32_t addrNum = 0;

		if (addr) { sscanf(addr, "%x", &addrNum);
		if (!addrNum) { fprintf(stderr,"Warning: Address '%s' is not valid hex\n", addr); }
		}
		makeImg (argv[2], argv[3], argv[4], addrNum ,cmdline);
		exit(0);

	}

	char *fileName = argv[1];
	int fd;
	int rc;
	int i;

	char *offsetArg = NULL;
	int offset = 0;
	int extract = 0;
	
	char *cmdline = NULL;
	char *addr = NULL;

	for (i = 1; i <= argc - 1; i++)
	{
	  if (strncasecmp(argv[i], "extract", 7) == 0) extract++;
	  if (strstr(argv[i], "offset=")) {
			   sscanf (argv[i] + strlen("offset=") ,"%d", &offset);
			}
	}

	char *mmapped = NULL;
	
	struct stat stbuf;

	rc = stat(fileName, &stbuf);

	if  (!(stbuf.st_mode & S_IFREG)) { fprintf (stderr,"%s is not a file\n", fileName); exit(6); }

	if (rc < 0) { perror(fileName); exit (5); }

 	fd = open (fileName, (cmdline? O_RDWR : O_RDONLY));
	if (fd < 0) { perror (fileName); exit(3);}

	
	mmapped =  mmap(NULL,
			stbuf.st_size,
			(cmdline? PROT_WRITE : PROT_READ),
			MAP_PRIVATE,
			fd,
			0);



	if (!mmapped) { fprintf (stderr, "IAH, we have a problem...\n"); perror ("mmap"); exit(4); }


	if (offset) mmapped += offset;

	if (strncmp(mmapped,BOOT_MAGIC, strlen(BOOT_MAGIC)) == 0)
	{
		// This is a boot image
		printf("Boot image detected\n");



		struct boot_img_hdr *bih = (struct boot_img_hdr *) mmapped;

		
		// Just like the header says:

		int n = (bih->kernel_size + bih->page_size - 1) / bih->page_size;
		int m = (bih->ramdisk_size + bih->page_size - 1) / bih->page_size;
		int o = (bih->second_size + bih->page_size - 1) / bih->page_size;


		printf ("Part      \tSize      \tPages\t  Addr\n");
		printf ("Kernel:   \t%-8d\t%-5d\t0x%x\n", bih->kernel_size, n,bih->kernel_addr);
		printf ("Ramdisk:  \t%-8d\t%-5d\t0x%x\n", bih->ramdisk_size, m, bih->ramdisk_addr);
		printf ("Secondary:\t%-8d\t%-5d\t0x%x\n", bih->second_size, o, bih->second_addr);
		printf ("Tags:       %x\n", bih->tags_addr);
		printf ("Flash Page Size: %d bytes\n", bih->page_size);
		printf ("DT Size: %d bytes\n", bih->dt_size);
		printf ("ID: ");
		int id =0;
		for (id = 0 ; id <8 ; id++)
		{
			printf("%x", bih->id[id]);
		}
		printf("\n");


		printf ("Name:    %s\n", bih->name);
		printf ("CmdLine: %s\n", bih->cmdline);


		if (cmdline)
		{
			printf ("Overriding cmdline with %s\n", cmdline);

			strcpy (bih->cmdline, cmdline);
			close(fd);
			fd = open (fileName, O_RDWR);
			mmapped -= offset;

			write (fd, mmapped, stbuf.st_size);
			close(fd);

			exit(0);
		}

		if (extract)
		{
			fprintf (stderr,"Extracting contents to directory: %s\n", EXTRACTED_DIR);
			mkdir (EXTRACTED_DIR, 0777); // yeah, I know. Insecure. Bleh


			decompressKernel (mmapped +  1 * bih->page_size, bih->kernel_size);
			extractPortion (mmapped, 1 * bih->page_size, bih->kernel_size, "kernel");
			extractPortion (mmapped, (n+1) * bih->page_size, bih->ramdisk_size, "ramdisk");
			extractPortion (mmapped, (n+1) * bih->page_size, bih->ramdisk_size, "ramdisk");
			if (bih->second_size)
			{
			  // secondary is often not included in image (secondsize == 0)

			extractPortion (mmapped, (n+1) * bih->page_size, bih->second_size, "second");

			}

		} // extract

		exit(0);
	}

	if ((*((uint32_t *) mmapped)) == SIMG_MAGIC)
	{
		// This is a sparse image

		struct simg_header *si = (struct simg_header *) mmapped;

		printf("Sparse image v%d.%d detected\n", si->version_major, si->version_minor);
		printf("%d blocks of %d bytes compressed into %d chunks (%lld%% compressed)\n",
			si->num_blocks, si->block_size, si->num_chunks,
			100 - (100 * stbuf.st_size / (si->num_blocks * si->block_size)) );


		if (extract) extractSimg(fd);
		exit(0);


	}
	if (strncmp(mmapped,BOOTLDR_MAGIC, strlen(BOOTLDR_MAGIC)) == 0)
	{
		// This is a bootldr image
		printf("Boot loader detected\n");
		struct bootloader_images_header *bih = (struct bootloader_images_header *) mmapped;
		printf("%d images detected, starting at offset 0x%x. Size: %d bytes\n",
			bih->num_images,
			bih->start_offset,
			bih->bootldr_size);

		int total = 0;

		for (i = 0;
		     i <  bih->num_images;
		     i++)
		{
			printf ("Image: %d\tSize: %7d bytes\t%s\n",
				i,
				bih->img_info[i].size,
				bih->img_info[i].name);
			if (extract)
			{
				
				extractPortion (mmapped, bih->start_offset + total, bih->img_info[i].size,  bih->img_info[i].name);


			}

			total += bih->img_info[i].size;

		}


		if (total != bih->bootldr_size) fprintf (stderr,"Warning: Total mismatches reported bootloader size\n");

	

		
		exit(0);

	}
	
	
	if (strncmp(mmapped,IMGDATA_MAGIC, strlen(IMGDATA_MAGIC)) == 0)
	{
		printf("Image Data detected\n");
		struct imgdata_header *idh = (struct imgdata_header *) mmapped;
		printf("%d images detected, %d, %d %d\n",
			idh->num_images,
			idh->unknown,
			idh->unknown1,
			idh->unknown2);

		for (i = 0;
		     i <  idh->num_images;
		     i++)
		{
			printf ("Image: %d\tName: %16s\tDimensions: %5dx%-5d Unknowns: %5d %5d Offset: %7x Size: %7x\n",
				i,
				idh->img_info[i].name,
				idh->img_info[i].height,
				idh->img_info[i].width,
				idh->img_info[i].unknown3,
				idh->img_info[i].unknown4,
				idh->img_info[i].flash_offset,
				idh->img_info[i].flash_size);

			extractPortion (mmapped, idh->img_info[i].flash_offset, idh->img_info[i].flash_size,  idh->img_info[i].name);

		}
		exit(0);

	}

	fprintf (stderr,"%s is not a recognized image. Sorry\n", fileName);


}

